/* GET BATTERY STATUS VIA IO INSTRUCTIONS TO CONTROLLER
 * adapted from Hitachi diagnostic code
 */
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <asm/io.h>
#include "batt.h"

/* Definition for ALI M1535 SM-Bus Access I/O BASE address
 */
#define SMBBASE  0x8040

#define bool(v)	((v) ? 1 : 0)

int dm_timeout = {0};

/* published externals
 */
int ac_exist = {0};
int battery_exist = {0};
int charge_status = {0};
int battery_capacity = {0};

#if 0
int main(int argc, char *argv[])
	{
	get_batt_data();
	printf("ac present %d battery present %d charge status %d battery capacity %d\n", ac_exist, battery_exist, charge_status, battery_capacity);
	}
#endif

/* Get & AC Battery Data from Quanta Power Control Unit
 */
BINFO *get_batt_data(BINFO *p)
	{
	extern unsigned int smbus_w_read();
	unsigned int pcu_data_2f = {0xffff};
	unsigned int pcu_data_36 = {0xffff};

	if (iopl(3))
    	{
		perror("can't access I/O");
		exit(2);
		} 
	if ((pcu_data_2f = smbus_w_read(0x14, 0x2f)) != 0xffff)
		{
		if (!bool(ac_exist = pcu_data_2f & 0x0002))
			charge_status = 0;
		charge_status = bool(pcu_data_2f & 0x0010);
		if (pcu_data_2f & 0x8000)
			{
			battery_exist = pcu_data_36 = smbus_w_read(0x14, 0x36);
			battery_capacity = (pcu_data_2f & 0x7f00) >> 8;
			}
		else
			{
			battery_exist = charge_status = 0;
			battery_capacity = -1;
			}
		}
	p->s = pcu_data_2f;
	if (pcu_data_36 == 3600)
		p->b.size = BINFO_LARGE;
	else if (pcu_data_36 == 1800)
		p->b.size = BINFO_SMALL;
	else
		p->b.size = BINFO_UNKNOWN;
	iopl(0);
	return (p);
	}

void deadman()
	{
	extern char *procname;

	++dm_timeout;
	fprintf(stderr, "%s: hung in smbus_w_read(), resetting sm-bus\n", procname);
	}

/* Ali M1535 Chipset PMU Sm-Bus Command Routine -- returns word data
 */
unsigned int smbus_w_read(unsigned char address, unsigned char command)
	{
	void clear_all_status();
	void io_delay();
	unsigned char val;
	unsigned int return_data, i;

	address = address | 0x01; /* read operation, set bit 0 */
	clear_all_status();	/* SMB_STS clear bit 3-7, set bit 2 */

	/*Output IO port*/
	outb(address, SMBBASE+3);     /*DEV_ADDR*/
	io_delay();
	outb(command, SMBBASE+7);     /*SMB_CMD*/
	io_delay();
	outb(0x30, SMBBASE+1);        /*SMB_TYP*/
	io_delay();
	outb(0x00, SMBBASE+4);        /*SMB_DATA0*/
	io_delay();
	outb(0x00, SMBBASE+5);        /*SMB_DATA1*/
	io_delay();

	/* Generate SM-Bus Command to Bus */
	outb(0x00, SMBBASE+2);        /*STR_PRT*/
	io_delay();

	alarm(2);
	dm_timeout = 0;
	signal(SIGALRM, deadman);
	for (;;)	/* wait until status set */
		{
		val = (unsigned char)inb(SMBBASE+0);
		io_delay();
		if (val & 0xf0)	/* wait until Bit4-7 != 0 */
			break;
		if (dm_timeout)
			{
			outb(0x04, SMBBASE+1);
			sleep(1);
			return (-1);
			}
		}
	alarm(0);
	if (!(val & 0xe0))
		{
		return_data = (unsigned int)inb(SMBBASE+5);
		io_delay();
		return_data = return_data << 8;
		return_data = return_data|((unsigned int)inb(SMBBASE+4));
		io_delay();
		return (return_data); 
		}
	else
		return (0xffff); 
	}

/* Ali1535 SM-Bus Status Clear Routine
 */
void clear_all_status()
	{
	void io_delay();
	unsigned char val;

	while (1)
		{	/* wait until Bit3-7 == 0 */
		val = (unsigned char)inb(SMBBASE+0);
		io_delay();
		if (val & 0xf8)
			{
			outb(val,SMBBASE+0);
			io_delay();
			}
		else
			break;
		}
	while (1)
		{	/* wait until Bit2 == 1 */
		val = (unsigned char)inb(SMBBASE+0);
		io_delay();
		if (val & 0x04)
			break;
		}
	}

/* General I/O Delay Function
 */
void io_delay()
	{
  	outb(0x00,0xED);	/* Output Dummy Data to I/O port EDh */
	}
