/* POWER MANAGEMENT DAEMON BACKLIGHT AUTO-OFF
 * Copyright (C) 2000  Monta Vista Software Inc.
 *
 * Inspite of the fact X11 manages to consume far, far more than
 * its worth of scarce system resources, and will allow you to do
 * most anything from mostly anywhere, there STILL IS NO CLEAN,
 * SIMPLE, STANDARD METHOD TO DETERMINE THE LENGTH OF TIME OF USER
 * INPUT INACTIVITY.
 *
 * The following sad story is testament to this fact and is packaged
 * here as a library because: I hate it and don't want to look at it
 * anymore, I hope noone else will see it, and I hope this will all
 * go away like a bad dream at some point when X GETS FIXED.
 *
 *	Revision History
 *
 *	2000.11.15	Created.	john@mvista.com
 *	2001.01.30	Changed to use /proc/pm_interrupts.	john@mvista.com
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include <signal.h>
#include <stdio.h>
#include "pm.h"

/* how much of the start of each line we need
 */
#define INTFILE	"/proc/pm_interrupts"
#define INTBUFLEN 100	/* max interrupt file line length */
#define INTPFXLEN 20	/* read only minimally needed prefix */

#define BACKLIGHT_UTIL "/bin/backlight"

/* those levels which we will monitor
 */
typedef struct {
	char *substr;
	int enable;
	} INTMON;
#ifdef PROTOTYPE
INTMON ilevels[] = {
	{"keyboard", PMVAR_BLE},
	{"Mouse", PMVAR_BLE},
	{NULL, 0}};
#else
INTMON ilevels[] = {
	{"mk712", PMVAR_BLE},
	{"usb-ohci", PMVAR_BLE},
	{NULL, 0}};
#endif

#define NILEVELS (sizeof (ilevels) / sizeof (INTMON) - 1)
char a_snap[NILEVELS][INTPFXLEN + 1];
char b_snap[NILEVELS][INTPFXLEN + 1];

/* state variables
 */
static int bl_power_on = {1};
static int to_limit = {0};
static int to_count = {0};

/* pmd needs to see this
 */
int bl_state = {0};

/* timer control
 */
static struct itimerval itv_off = {0, 0, 0, 0};
static struct itimerval itv_one_sec = {1, 0, 1, 0};
#define stop_timer()	set_timer(&itv_off)
#define start_timer()	set_timer(&itv_one_sec)

/* local functions
 */
static void timeout();
static char *instr();
static int read_c();
static int cmpsnap();
static int get_ilevels();

/* do whatever is needed to turn on (arg) or off (!arg) backlight
 */
int bl_power(arg)
	{
	char buf[80];

#ifdef PROTOTYPE
	printf("Power %s\n", arg ? "ON" : "OFF");
#else
	sprintf(buf, "%s %s\n", BACKLIGHT_UTIL, arg ? "on" : "off");
	system(buf);
#endif
	bl_power_on = arg;
	}

#ifdef TEST_THIS_MESS
main()
	{

	set_backlight_to(5, bl_power);
	for (;;)
		sleep(1);
	}
#endif

int set_timer(struct itimerval *time)
	{
	if (setitimer(ITIMER_REAL, time, NULL))
		{
		perror("can't set timer");
		exit (2);
		}
	}

/* set backlight time/power state
 */
int set_bl_state(int state, int sec)
	{
	static int toggle = {0};
	static int settle = {0};
	int do_timer;
	int mask;
	int inactive;
int is = state;

	if (bl_state == BL_SHUTDOWN)    /* terminal state */
		return;
	mask = sigblock(sigmask(SIGALRM));
	switch (state)
		{
	case BL_SHUTDOWN:
		break;
	case BL_UNTIMED_ON:
		bl_power(1);
		to_limit = to_count = 0;
		break;
	case BL_TIMED_ON:
		bl_power(1);
	case BL_NEWTIME:
		if (!(to_limit = sec))
			state = BL_UNTIMED_ON;	/* degenerate case */
		to_count = 0;
		break;
	case BL_OFFEWAIT:
		bl_power(0);
		break;
	case BL_USREVENT:
		to_count = 0;
		bl_power(1);
		state = to_limit ? BL_TIMED_ON : BL_UNTIMED_ON;
		break;
	case BL_SETTLE:
		settle = 1;
		state = bl_state;
		break;
	case BL_TICK:
		state = bl_state;	/* stay in same state by default */
		if (bl_state != BL_UNTIMED_ON)
			{
			++toggle;
			get_ilevels(toggle & 1 ? a_snap : b_snap);
			inactive = cmpsnap();
			if (!inactive && !settle)
				{
				to_count = 0;
				if (!bl_power_on)
					{
					bl_power(1);
					state = to_limit ? BL_TIMED_ON : BL_UNTIMED_ON;
					}
				}
			else if (inactive && settle)
				{	/* input has quieted */
				settle = 0;
				bl_power(0);
				state = BL_OFFEWAIT;
				}
			else if (!inactive && settle)
				;	/* wait for input to quiet */
			else if (!bl_power_on)
				;
			else if (to_limit <= ++to_count)
				{
				bl_power(0);
				state = BL_OFFEWAIT;
				}
			}
		break;
	case BL_INIT:
		signal(SIGALRM, timeout);
		start_timer();
		bl_power(1);
		to_limit = to_count = 0;
		state = BL_UNTIMED_ON;
		break;
	default:
		fprintf(stderr, "set_bl_state: unknown state: %d\n", state);
		state = bl_state;
		}
	bl_state = state;
	sigsetmask(mask);
#ifdef DEBUG
printf("set_bl_time: sec %d is %d state %d tol %d toc %d bl_pwr_on %d\n",
sec, is, state, to_limit, to_count, bl_power_on);
#endif
	}

/* pull-in lines from /proc/pm_interrupts which match the ilevels strings
 */
static int get_ilevels(char snap[NILEVELS][INTPFXLEN + 1])
	{
	extern PMVARS cur_vars[];
	INTMON *pim;
	char *p;
	int nl;
	int c;

	read_c(1);
	for (nl = 0; nl < NILEVELS; )
		{
		for (p = snap[nl]; ; )
			if ((c = read_c(0)) == -1)
				return (*p = '\0', -1);
			else if (c == '\n')
				{
				*p = '\0';
				break;
				}
			else if (INTPFXLEN < p - snap[nl])
				;	/* silently drop */
			else
				*p++ = c;
#if 0
		printf("read: [%s]\n", snap[nl]);
#endif
		for (pim = ilevels; pim->substr; ++pim)
			if (cur_vars[pim->enable].data && instr(snap[nl], pim->substr))
				{
#if 0
printf("found [%s]0x%x\n", snap[nl], snap[nl]);
#endif
				++nl;
				break;
				}
		}
	return (nl);
	}

/* search for *ss in *str, return start address of pattern
 */
static char *instr(char *str, char *ss)
	{
	char *ostr = str;
	char *p, *q, *r;

	for (p = ss; ; ++str)
		if (!*str)
			return (0);
		else if (*str != *p)
			;
		else
			for (q = p, r = str; ; ++q, ++r)
				if (!*q)
					return (str);
				else if (*q != *r)
					break;
	}
					
/* read /proc/pm_interrupts in gulps of INTBUFLEN, rewind file if requested
 */
static int read_c(int rewind)
	{
	static int fd = {0};
	static int nc = {0};
	static int pt = {0};
	static char buf[INTBUFLEN];

	if (!fd && (fd = open(INTFILE, O_RDONLY)) < 0)
		{
		perror("can't open: "INTFILE);
		exit(2);
		}
	if (rewind)
		lseek(fd, 0, SEEK_SET);
	else
		for (;;)
			if (nc != pt)
				return (buf[nc++]);
			else if (nc = 0, (pt = read(fd, buf, INTBUFLEN)) < 0)
				return (perror("read error"), -1);
			else if (!pt)
				return (-1);	/* EOF */
	}

/* SIGALRM handler, manages timeout/auto-on state
 */
static void timeout()
	{
	set_bl_state(BL_TICK, 0);
	}

/* compare current and previous interrupt tally snapshots,
 * return true if equal
 */
static int cmpsnap()
	{
	char *p, *q;
	int i;

	for (i = 0; i < NILEVELS; ++i)
		for (p = a_snap[i], q = b_snap[i]; ; ++p, ++q)
			if (*p != *q)
				return (0);
			else if (!*p)
				break;
	return (1);
	}
