#include <stdio.h>
#include <unistd.h>
#include <asm/io.h>
#include <signal.h>

#define ON 1
#define OFF 0

/* Definition for ALI M1535 SM-Bus Access I/O BASE address
 */
#define SMBBASE  0x8040

#define bool(v)	((v) ? 1 : 0)

/* published externals
 */
int ac_exist = {0};
int battery_exist = {0};
int charge_status = {0};
int battery_capacity = {0};

#define CHECKPOINT(n, d)	(cp_name = (n), cp_data = (d))
char *cp_name;
unsigned cp_data;

/* Read Data from PCU */
      /*  PCU Slave Address = 14h                            */
      /*   AC Battery Status Command = 2fh                   */
      /*                       Bit1 = 1:AC Exist             */
      /*                       Bit4 = 1:Charging             */
      /*                       Bit15= 1:Battery Exist        */
      /*                       Bit8-14 = Battery Capacity    */
      /*   Battery Capacity Command = 36h                    */
      /*                                                     */

union {
	struct {
		unsigned int x0:1;
		unsigned int ac:1;
		unsigned int x1:2;
		unsigned int charge:1;
		unsigned int x2:3;
		unsigned int capacity:7;
		unsigned int exist:1;
		} b;
	unsigned short s;
	} battery;

void checkpoint()
	{
	fprintf(stderr, "hung at point %s data 0x%x\n", cp_name, cp_data);
	_exit(0);
	}

#if 1
int main(int argc, char *argv[])
	{
	signal(SIGALRM, checkpoint);
	alarm(5);
	get_data();
	printf("ac present %d battery present %d charge status %d battery capacity %d\n", ac_exist, battery_exist, charge_status, battery_capacity);
	}
#endif

/* Get & AC Battery Data from Quanta PCU(Power Control Unit)
 */
int get_data()
	{
	extern unsigned int smbus_w_read();

	unsigned int pcu_data_2f = {0xffff};
	unsigned int pcu_data_36 = {0xffff};
	unsigned int buffer = {0x0000};

	if (iopl(3))
    		{
		perror("can't access I/O");
		exit(2);
		} 
	if ((pcu_data_2f = smbus_w_read(0x14, 0x2f)) != 0xffff)
		{
		if (!bool(ac_exist = pcu_data_2f & 0x0002))
			charge_status = 0;
		charge_status = bool(pcu_data_2f & 0x0010);
		if (pcu_data_2f & 0x8000)
			{
			battery_exist = pcu_data_36 = smbus_w_read(0x14, 0x36);
			battery_capacity = (pcu_data_2f & 0x7f00) >> 8;
			}
		else
			{
			battery_exist = charge_status = 0;
			battery_capacity = -1;
			}
		}
	iopl(0);
	}

/* Ali M1535 Chipset PMU Sm-Bus Command Routine [Read Word] 
 */
unsigned int smbus_w_read(unsigned char address, unsigned char command)
	{
	void clear_all_status();
	void io_delay();
	unsigned char buffer;
	unsigned int return_data, i;

	address = address | 0x01; /* read operation, set bit 0 */
CHECKPOINT("smbus_w_read: enter", 0); 
	clear_all_status();	/* SMB_STS clear bit 3-7, set bit 2 */
CHECKPOINT("smbus_w_read: after clear_all_status", 0);

	/*Output IO port*/
	outb(address,SMBBASE+3);     /*DEV_ADDR*/
	io_delay();
	outb(command,SMBBASE+7);     /*SMB_CMD*/
	io_delay();
	outb(0x30,SMBBASE+1);        /*SMB_TYP*/
	io_delay();
	outb(0x00,SMBBASE+4);        /*SMB_DATA0*/
	io_delay();
	outb(0x00,SMBBASE+5);        /*SMB_DATA1*/
	io_delay();

	/* Generate SM-Bus Command to Bus */
	outb(0x00,SMBBASE+2);        /*STR_PRT*/
	io_delay();

	while(1)	/* wait until status set */
		{
		buffer = (unsigned char)inb(SMBBASE+0);
CHECKPOINT("smbus_w_read: wait status set", buffer);
		io_delay();
		if (buffer & 0xf0)	/* wait until Bit4-7 != 0 */
			break;
		}
	if (!(buffer & 0xe0))
		{
		return_data=(unsigned int)inb(SMBBASE+5);
		io_delay();
		return_data = return_data << 8;
		return_data = return_data|((unsigned int)inb(SMBBASE+4));
		io_delay();
		return (return_data); 
		}
	else
		return (0xffff); 
CHECKPOINT("smbus_w_read: returns", 0);
	}

/* Ali1535 SM-Bus Status Clear Routine
 */
void clear_all_status()
	{
	void io_delay();
	unsigned char buffer;

	while (1)
		{	/* wait until Bit3-7 == 0 */
		buffer = (unsigned char)inb(SMBBASE+0);
CHECKPOINT("clear_all_status: wait for zero", buffer);
		io_delay();
		if (buffer & 0xf8)
			{
			outb(buffer,SMBBASE+0);
			io_delay();
			}
		else
			break;
		}
	while (1)
		{	/* wait until Bit2 == 1 */
		buffer = (unsigned char)inb(SMBBASE+0);
CHECKPOINT("clear_all_status: wait for one", buffer);
		io_delay();
		if (buffer & 0x04)
			break;
		}
	}

/* General I/O Delay Function
 */
void io_delay()
	{
  	outb(0x00,0xED);	/* Output Dummy Data to I/O port EDh */
	}
