/* POWER MANAGEMENT USER INTERFACE
 * Copyright (C) 2000  Monta Vista Software Inc.
 */
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <sys/file.h>
#include <gtk/gtk.h>
#include "pm.h"
#include "pmvars.h"
#include "pmlocale.h"

/* internal defs
 */
#define SETVARS(var, val)	(cur_vars_mod = 1, cur_vars[var].data = val)

/* labeled status bar
 */
typedef struct {
	GtkWidget *ac_input;
	GtkWidget *charge_state;
	GtkWidget *label;
	GtkWidget *pb;
	} LPB;

PMVARS cur_vars[NVARS] = {0};
PMVARS org_vars[NVARS] = {0};
int cur_vars_mod = {0};

/* forward refs
 */
LPB *update_lpb(LPB *, gfloat, int, int);

/* command lime flags
 */
int verbfl = {0};

/* setup current variables from default template
 */
unsigned int init_vars()
	{
	PMVARS *pv, *qv;

	for (pv = cur_vars, qv = def_vars; (pv->data = qv->data,
			(pv++)->name = (qv++)->name); )
				;
	return (0);
	}

/* generate a new, full configuration file
 */
char *write_vars(char *fn, PMVARS *pv)
	{
	FILE *file;

	if (!(file = fopen(fn, "w")))
		return ("Can't write configuration file");
	for ( ; pv->name; ++pv)
		fprintf(file, "%s %d\n", pv->name, pv->data);
	fclose(file);
	return (NULL);
	}

/* set vars from config file, vars may be in arbitrary order
 * O(n^2) algorithm, but n should be small
 */
unsigned int read_vars(char *fn)
	{
	FILE *file;
	char name[100];	/* see fscanf %s arg below */
	PMVARS *pv, *qv;
	unsigned int val;

	if (!(file = fopen(fn, "r")))
		{
		perror("Can't open configuration file");
		exit(2);
		}
	while (fscanf(file, "%99s %d", name, &val) == 2)
		for (pv = cur_vars; ; ++pv)
			if (!pv->name)
				{
				fprintf(stderr,
					"read_vars: ignoring unknown parameter: \"%s\"\n", name);
				break;
				}
			else if (!strcmp(pv->name, name))
				{
#if 0
g_print("read_vars: found %s setting data as %d was %d\n",
name, val, pv->data);
#endif
				pv->data = val;
				break;
				}
	for (pv = org_vars, qv = cur_vars; (pv->data = qv->data,
			(pv++)->name = (qv++)->name); )
				;
	fclose(file);
	}

/* monitor battery state, update level progress bar
 */
gint batt_mon(gpointer lpb)
	{
	static int err_cnt = {0};
	GtkWidget *pb = ((LPB *)lpb)->pb;
	GtkAdjustment *adj;
	int battery_level, ac_input, charging, battery_present, val;
	char name[100];	/* see fscanf %s arg below */
	FILE *file;
	char *err;


	err = NULL;
	if (!(file = fopen(POWER_STATUS_PATH, "r")))
		err = "open";
	else if (flock(file->_fileno, LOCK_SH|LOCK_NB))
		err = "lock";
	else
		while (fscanf(file, "%99s %d", name, &val) == 2)
			if (!strcmp("battery_level", name))
				battery_level = val;
			else if (!strcmp("ac_input", name))
				ac_input = val;
			else if (!strcmp("charging", name))
				charging = val;
			else if (!strcmp("battery_present", name))
				battery_present = val;
	if (file)
		{
		flock(file->_fileno, LOCK_UN);
		fclose(file);
		}
	if (!err)
		{
		err_cnt = 0;
		adj = GTK_PROGRESS(pb)->adjustment;
		if (battery_level < adj->lower)
			battery_level = adj->lower;
		else if (adj->upper < battery_level)
			battery_level = adj->lower;
		update_lpb((LPB *)(lpb), battery_present ? (gfloat)battery_level : -1,
			ac_input, charging);
		}
	else if (++err_cnt < 3)
		;
	else
		fprintf(stderr, "Daemon status file error: %s\n", err);
	return (TRUE);
	}

/* update label and progress button, creat if !lpb
 */
LPB *update_lpb(LPB *lpb, gfloat pct, int ac_input, int charging)
	{
	char a[100];
	char b[100];
	char c[100];
	int bp;

	sprintf(b, "%s", ac_input ? LC_AC_ADAPTOR : LC_BATTERY);
	if (pct < 0)
		pct = bp = 0;
	else
		bp = 1;
	sprintf(a, LC_REMAINING_BATTERY_CAPACITY": %3d%%\n", (int)pct);
	sprintf(c, "%s", !bp ? LC_NOT_PRESENT : (!ac_input ? LC_DISCHARGING :
		(charging ? LC_CHARGING : LC_CHARGE_OFF)));
	if (!lpb)
		{
		lpb = (void *)malloc(sizeof (LPB));
		lpb->pb = gtk_progress_bar_new();
		lpb->label = gtk_label_new(a);
		lpb->ac_input = gtk_label_new(b);
		lpb->charge_state = gtk_label_new(c);
		}
	else
		{
		gtk_label_set_text((GtkLabel *)lpb->label, a);
		gtk_label_set_text((GtkLabel *)lpb->ac_input, b);
		gtk_label_set_text((GtkLabel *)lpb->charge_state, c);
		}
	gtk_progress_set_value(GTK_PROGRESS(lpb->pb), pct);
	return (lpb);
	}

/* wack the pm daemon when the conf file changes
 */
char *signal_daemon(char *fn, int sig)
	{
	FILE *file;
	unsigned int pid;

	if (!(file = fopen(fn, "r")))
		return ("can't open daemon pid file");
	else if (fscanf(file, "%d", &pid) != 1)
		return ("corrupt daemon pid file");
	else if (kill(pid, sig) < 0)
		return ("can't signal daemon");
	return (NULL);
	}

/* exit pm utility
 */
gint delete(GtkWidget *window, GtkButton *button)
	{
	gtk_main_quit();
	return (FALSE);
	}

/* power management item enable button handler
 */
void pm_enable(GtkButton *button, guint info)
	{
	SETVARS(info, GTK_TOGGLE_BUTTON(button)->active);
	}

/* power management item timeout range handler
 */ 
void pm_timeout(GtkAdjustment *adj, guint info)
	{
	SETVARS(info, (guint)adj->value * SEC_MIN);
	}

/* power button mode selector (radio button) handler
 */
void pb_mode(GtkButton *button, guint mode)
	{
	SETVARS(PMVAR_MD, mode);
	}

/* simulated power button, button handler
 */
void power_button(GtkButton *button, guint info)
	{
	signal_daemon(PMD_PID_PATH, SIGUSR1);
	}

/* pack labelled bottons into vertical box
 */
GtkWidget *bt_entry(GtkWidget *vb, char *name, guint idx)
	{
	GtkWidget *hb, *checkbutton;

	hb = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vb), hb, FALSE, TRUE, 10);
	checkbutton = gtk_check_button_new_with_label(name);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton),
		cur_vars[idx].data);
	gtk_widget_set_usize(checkbutton, 200, 20);
	gtk_box_pack_start(GTK_BOX(hb), checkbutton, FALSE, TRUE, 5);
	gtk_signal_connect(GTK_OBJECT(checkbutton), "clicked",
		(GtkSignalFunc)pm_enable, (gpointer)idx);
	return (hb);
	}

/* make an item entry in power mgmt window
 */
GtkWidget *pm_entry(GtkWidget *vb, char *name, guint enb_idx, guint t_idx,
	int enabled)
	{
	GtkWidget *hb, *checkbutton, *spinner, *label;
	GtkAdjustment *adj;

	hb = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vb), hb, FALSE, TRUE, 10);
	checkbutton = gtk_check_button_new_with_label(name);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton),
		cur_vars[enb_idx].data);
	gtk_widget_set_sensitive(checkbutton, enabled);
	gtk_widget_set_usize(checkbutton, 225, 20);						// hitachi mod
	gtk_box_pack_start(GTK_BOX(hb), checkbutton, FALSE, TRUE, 5);
	gtk_signal_connect(GTK_OBJECT(checkbutton), "clicked",
		(GtkSignalFunc)pm_enable, (gpointer)enb_idx);
	adj = (GtkAdjustment *)gtk_adjustment_new(10, 1, 60, 1, 5, 0);
	gtk_adjustment_set_value(adj, (gfloat)cur_vars[t_idx].data / SEC_MIN);
	spinner = gtk_spin_button_new(adj, 0, 0);
	gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(spinner), TRUE);
	gtk_spin_button_set_shadow_type(GTK_SPIN_BUTTON(spinner),
		GTK_SHADOW_OUT);
	gtk_signal_connect(GTK_OBJECT(adj), "value_changed",
		GTK_SIGNAL_FUNC(pm_timeout), (gpointer)t_idx); 
	gtk_box_pack_start(GTK_BOX(hb), spinner, FALSE, TRUE, 0);
	label = gtk_label_new(LC_MINUTES);
	gtk_box_pack_start(GTK_BOX(hb), label, FALSE, TRUE, 0);
	return (hb);
	}

/* create power mgmt page
 */
GtkWidget *pm_page(GtkWidget *notebook)
	{
	GtkWidget *frame = gtk_frame_new(LC_POWER_SAVING_MODE_TIMEOUTS);
	GtkWidget *vb, *label;
 
	gtk_container_set_border_width(GTK_CONTAINER(frame), 10);
	vb = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(frame), vb);
	pm_entry(vb, LC_BACKLIGHT_POWER_OFF, PMVAR_BLE, PMVAR_BL, TRUE);
	pm_entry(vb, LC_PCI_LAN_MODEM_POWER_OFF, PMVAR_LME, PMVAR_LM, FALSE);
	pm_entry(vb, LC_USB_POWER_OFF, PMVAR_POE, PMVAR_PO, FALSE);
	pm_entry(vb, LC_PCMCIA_POWER_OFF, PMVAR_PCE, PMVAR_PC, FALSE);
	gtk_widget_show_all(frame);
	label = gtk_label_new(LC_POWER_MANAGEMENT);
	gtk_notebook_insert_page(GTK_NOTEBOOK(notebook), frame, label, 0);
	}

/* create power status page
 */
GtkWidget *st_page(GtkWidget *notebook)
	{
	GtkWidget *frame = gtk_frame_new(LC_POWER_STATUS);
	GtkWidget *vb, *hb, *label, *pb, *align;
	char buf[100];
	LPB *lpb;
	lpb = update_lpb(NULL, (gfloat)50, 1, 1);
 
	gtk_container_set_border_width(GTK_CONTAINER(frame), 10);
	vb = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(frame), vb);

	hb = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vb), hb, FALSE, TRUE, 5);
	label = gtk_label_new(LC_POWER_SOURCE);
	gtk_box_pack_start(GTK_BOX(hb), label, FALSE, TRUE, 5);
	gtk_box_pack_start(GTK_BOX(hb), lpb->ac_input, FALSE, TRUE, 5);

	hb = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vb), hb, FALSE, TRUE, 5);
	label = gtk_label_new(LC_BATTERY_STATE);
	gtk_box_pack_start(GTK_BOX(hb), label, FALSE, TRUE, 5);
	gtk_box_pack_start(GTK_BOX(hb), lpb->charge_state, FALSE, TRUE, 5);

	gtk_box_pack_start(GTK_BOX(vb), lpb->label, FALSE, TRUE, 5);
	align = gtk_alignment_new(0.5, 0.5, 0, 0);
	gtk_box_pack_start(GTK_BOX(vb), align, FALSE, FALSE, 5);
	gtk_container_add(GTK_CONTAINER(align), lpb->pb); 
	gtk_widget_show_all(frame);
	label = gtk_label_new(LC_POWER_METER);
	gtk_notebook_insert_page(GTK_NOTEBOOK(notebook), frame, label, 0);
	gtk_timeout_add(BATT_POLL_INTV, batt_mon, lpb); 
	}

/* create alarm page
 */
GtkWidget *al_page(GtkWidget *notebook)
	{
	GtkWidget *vb0, *vb1, *hb, *label;
	GtkWidget *frame;
 
	vb0 = gtk_vbox_new(FALSE, 0);
	frame = gtk_frame_new(LC_LOW_BATTERY_NOTIFICATION);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 10);
	gtk_container_add(GTK_CONTAINER(vb0), frame);
	vb1 = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(frame), vb1);
	bt_entry(vb1, LC_AUDIBLE_WARNING, PMVAR_AU);
	bt_entry(vb1, LC_POP_UP_MESSAGE, PMVAR_PU);
	frame = gtk_frame_new(LC_ALARM_ACTIONS);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 10);
	gtk_container_add(GTK_CONTAINER(vb0), frame);
	vb1 = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(frame), vb1);
	bt_entry(vb1, LC_SHUTDOWN_SYSTEM, PMVAR_AC);
	gtk_widget_show_all(vb0);
	label = gtk_label_new(LC_POWER_ALARM);
	gtk_notebook_insert_page(GTK_NOTEBOOK(notebook), vb0, label, 0);
	}

/* create a group of radio buttons for power button mode
 */
GtkWidget *pbm_add_button(GtkWidget *vb, char *label, guint mode, GtkWidget *pb)
	{
	GtkWidget *button;
	GSList *group = pb ? gtk_radio_button_group(GTK_RADIO_BUTTON(pb)) : NULL;

	button = gtk_radio_button_new_with_label(group, label);
	gtk_widget_set_usize(button, 200, 20);
	gtk_box_pack_start(GTK_BOX(vb), button, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)pb_mode, (gpointer)mode);
	if (cur_vars[PMVAR_MD].data == mode)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
	return (button);
	}

/* create power button mode page
 */
GtkWidget *pb_page(GtkWidget *notebook)
	{
	GtkWidget *vb, *vb2, *button, *frame, *label;
	GSList *group;

	vb = gtk_vbox_new(FALSE, 10);
	gtk_container_set_border_width(GTK_CONTAINER(vb), 10);

	frame = gtk_frame_new(LC_POWER_BUTTON_MODE);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 10);
	gtk_box_pack_start(GTK_BOX(vb), frame, TRUE, TRUE, 0);

	vb2 = gtk_vbox_new(FALSE, 10);
	gtk_container_set_border_width(GTK_CONTAINER(vb2), 10);
	gtk_container_add(GTK_CONTAINER(frame), vb2);

	button = pbm_add_button(vb2, LC_ENTER_LOW_POWER_MODE, PBM_PS, NULL);
	button = pbm_add_button(vb2, LC_IMMEDIATE_POWER_OFF, PBM_PD, button);
	pbm_add_button(vb2, LC_PROMPT_USER_FOR_CONFIRMATION, PBM_PU, button);

#if 0
	button = gtk_button_new();
	label = gtk_label_new(LC_SIMULATE_POWER_BUTTON);
	gtk_container_add(GTK_CONTAINER(button), label);
	gtk_box_pack_start(GTK_BOX(vb), button, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(power_button), 0);
#endif

	label = gtk_label_new(LC_POWER_BUTTON);
	gtk_notebook_insert_page(GTK_NOTEBOOK(notebook), vb, label, 0);
	gtk_widget_show_all(vb);
	}

gint info_delete(GtkWidget *widget, GtkWidget *event, gpointer data)
	{
	gtk_widget_destroy((GtkWidget *)data);
	return (TRUE);
	}

gint info_close(GtkButton *button, gpointer data)
	{
	gtk_widget_destroy((GtkWidget *)data);
	return (TRUE);
	}

void info_popup(char *info)
	{
	GtkWidget *window, *label, *vb, *checkbutton;

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), LC_TITLE);					// hitachi add
	gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);	// hitachi add
	gtk_signal_connect(GTK_OBJECT(window), "delete_event",
			GTK_SIGNAL_FUNC(info_delete), window);
	gtk_container_set_border_width(GTK_CONTAINER(window), 10);
	vb = gtk_vbox_new(FALSE, 10);
	gtk_container_set_border_width(GTK_CONTAINER(vb), 10);
	gtk_container_add(GTK_CONTAINER(window), vb);
	label = gtk_label_new(info);
	gtk_box_pack_start(GTK_BOX(vb), label, FALSE, FALSE, 5);
	checkbutton = gtk_button_new_with_label(LC_OK);						// hitachi mod
	gtk_signal_connect(GTK_OBJECT(checkbutton), "clicked",
		(GtkSignalFunc)info_close, GTK_OBJECT(window));
	gtk_box_pack_start(GTK_BOX(vb), checkbutton, FALSE, FALSE, 5);
	gtk_widget_show_all(window);
	}

/* update configuration file and ping daemon
 */
int apply_button(GtkWidget *window, GtkButton *button)
	{
	char *msg;

	msg = NULL;

	if (!cur_vars_mod)
		;
	else if (msg = write_vars(PMCONFIG_PATH, cur_vars))
		;
	else if (msg = signal_daemon(PMD_PID_PATH, SIGHUP))
		;
	else
		{
		cur_vars_mod = 0;
		if (verbfl)
			msg = LC_SUCCESSFULLY_UPDATED_CONFIGURATION;
		}
	if (msg)
		info_popup(msg);
	}

/* update configuration and exit
 */
void ok_button(GtkWidget *window, GtkButton *button)
	{
	apply_button(window, button);
	delete(window, button);
	}

/* launch pmd if not present and start GUI
 */
int main(int argc, char *argv[])
	{
	GtkWidget *window;
	GtkWidget *button;
	GtkWidget *table;
	GtkWidget *notebook;
	int pmd_pid;
	
	if (!signal_daemon(PMD_PID_PATH, 0))
		;	/* daemon "should" already exist */
	else if (pmd_pid = fork())
		;
	else if (pmd_pid < 0)
		perror(LC_WARNING "couldn't fork, can't start: "PMD_PATH);
	else if (execl(PMD_PATH, PMD_PATH, NULL) < 0)
		perror(LC_WARNING "couldn't exec, can't start: "PMD_PATH);
	gtk_set_locale();
	gtk_init(&argc, &argv);
	init_vars();
	read_vars(PMCONFIG_PATH);
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), LC_TITLE);					// hitachi add
	gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);	// hitachi add
	gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, FALSE);		// hitachi add
	gtk_signal_connect(GTK_OBJECT(window), "delete_event",
		GTK_SIGNAL_FUNC(delete), NULL);
	gtk_container_set_border_width(GTK_CONTAINER(window), 10);
	table = gtk_table_new(3,6,FALSE);
	gtk_container_add(GTK_CONTAINER(window), table);
	notebook = gtk_notebook_new();
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook), GTK_POS_TOP);
	gtk_table_attach_defaults(GTK_TABLE(table), notebook, 0,6,0,1);
	gtk_widget_show(notebook);
	pb_page(notebook);
	pm_page(notebook);
	al_page(notebook);
	st_page(notebook);
	gtk_notebook_set_page(GTK_NOTEBOOK(notebook), 0);
	button = gtk_button_new_with_label("OK");
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)ok_button, GTK_OBJECT(window));
	gtk_table_attach_defaults(GTK_TABLE(table), button, 0,1,2,3);
	gtk_widget_show(button);
	button = gtk_button_new_with_label(LC_APPLY);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)apply_button, GTK_OBJECT(window));
	gtk_table_attach_defaults(GTK_TABLE(table), button, 1,2,2,3);
	gtk_widget_show(button);
	button = gtk_button_new_with_label(LC_CLOSE);
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(delete), GTK_OBJECT(window));
	gtk_table_attach_defaults(GTK_TABLE(table), button, 2,3,2,3);
	gtk_widget_show(button);
	gtk_widget_show(table);
	gtk_widget_show(window);
	cur_vars_mod = 0;	/* dismiss spurious updates */
	gtk_main();
	return (0);
	}
