/* POWER MANAGEMENT HEADER
 * Copyright (C) 2000  Monta Vista Software Inc.
 */

#ifndef __PM_H__
#define __PM_H__
/* system defaults
 */
#ifdef PROTOTYPE
#define PM_PATH		"./pm"
#define PMD_PATH	"./pmd"
#define PMACPID_PATH	"./pmacpid"
#define PMHALT_PATH	"./pmhalt"
#define PMCONFIG_PATH	"./pmconfig"
#define PMD_PID_PATH	"./pmd.pid"
#define POWER_STATUS_PATH "./power.status"
#define SEC_MIN 1
#endif

#ifdef PRODUCTION
#define PM_PATH		"/usr/local/bin/pm"
#define PMD_PATH	"/usr/local/bin/pmd"
#define PMACPID_PATH	"/usr/local/bin/pmacpid"
#define PMHALT_PATH	"/usr/local/bin/pmhalt"
#define PMCONFIG_PATH	"/home/hitachi/pm/pmconfig"		//need exist
#define PMD_PID_PATH	"/home/hitachi/pm/pmd.pid"		//auto create
#define POWER_STATUS_PATH "/home/hitachi/pm/power.status"	//auto create
#define SEC_MIN 60
#endif

#ifdef TEST_WEBPAD
#define PM_PATH		"./pm"
#define PMD_PATH	"./pmd"
#define PMACPID_PATH	"./pmacpid"
#define PMHALT_PATH	"./pmhalt"
#define PMCONFIG_PATH	"./pmconfig"
#define PMD_PID_PATH	"./pmd.pid"
#define POWER_STATUS_PATH "./power.status"
#define SEC_MIN 5
#endif

/* battery admin constants
 */
#define BATT_POLL_INTV	1000	/* 1 second */
#define LO_BATT_TRIP	5		/* 5% */

/* structure of configuration file variables
 */
typedef struct pmvars {
	char *name;
	unsigned int data;
	} PMVARS;

/* variable template offset indexes
 */
#define PMVAR_BLE	0	/* backlight enable:bool */
#define PMVAR_LME	1	/* pci lan/modem enable:bool */
#define PMVAR_POE	2	/* usb enable:bool */
#define PMVAR_PCE	3	/* pcmcia enable:bool */
#define PMVAR_AU	4	/* low batt audible warning:bool */
#define PMVAR_PU	5	/* low batt pop-up warning:bool */
#define PMVAR_AC	6	/* low batt shutdown:bool */
#define PMVAR_MD	7	/* power button mode:sel */
#define PMVAR_BL	8	/* backlight:min */
#define PMVAR_LM	9	/* pci lan/modem:min */
#define PMVAR_PO	10	/* usb:min */
#define PMVAR_PC	11	/* pcmcia:min */

/* PMVAR_MD data -- powerbutton mode
 */
#define PBM_PS	0	/* enter power saving mode */
#define PBM_PD	1	/* power down */
#define PBM_PU	2	/* prompt user */

/* possible states/commands for backlight control
 *
 *                          SET TO   MODIFY TIME PARAMS   TIMEOUT
 */
#define BL_UNTIMED_ON 1 /*  on       no                   off          */
#define BL_NEWTIME  2   /*  --       yes                  yes, restart */
#define BL_OFFEWAIT 3   /*  off      no                   --           */
#define BL_TIMED_ON 4   /*  on       yes                  yes, start   */

/* additional commands -- NB: these are not states
 */
#define BL_USREVENT	5	/* user input activity */
#define BL_TICK		6	/* timer event */
#define BL_INIT		7	/* initialize mechanism */
#define BL_SETTLE	8	/* wait for user activity to subside */

/* final state, once we get here we never get out
 */
#define BL_SHUTDOWN 9

#endif __PM_H__
