/* POWER MANAGEMENT DAEMON
 * Copyright (C) 2000  Monta Vista Software Inc.
 */
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <sys/file.h>
#include <gtk/gtk.h>
#include "pm.h"
#include "pmvars.h"
#include "pmlocale.h"

/* power buttom pop-up menu choices
 */
#define SD_IGN	0
#define SD_YES	1
#define SD_PS	2

/* daemon control
 */
#define DAEMON_START 1
#define DAEMON_DELETE 2

PMVARS cur_vars[NVARS] = {0};
char *procname = {NULL};
int lowbatt_warning = {0};
int lobatt_secs = {0};
int battery_level, ac_input, pb_pressed;

/* forward refs
 */
void power_monitor();
void cleanup();

/* local state variables
 */
int pb_popup_nest = {0}; /* prevent multiple power button pop-up windows */

/* setup current/new variables from default template
 */
unsigned int init_vars()
	{
	PMVARS *pv, *rv;

	for (pv = cur_vars, rv = def_vars;
		(pv->data = rv->data, (pv++)->name = (rv++)->name); )
			;
	return (0);
	}

/* dump current variable data
 */
void dump_vars()
	{
	PMVARS *pv;

	for (pv = cur_vars; pv->name; ++pv)
		g_print("%s: %d\n", pv->name, pv->data);
	}

/* set vars from config file, vars may be in arbitrary order
 * O(n^2) algorithm, but n should be small
 */
char *read_vars(char *fn)
	{
	FILE *file;
	char name[100];	/* see fscanf %s arg below */
	PMVARS *pv;
	unsigned int val;

	if (!(file = fopen(fn, "r")))
		return ("Can't open configuration file");
	while (fscanf(file, "%99s %d", name, &val) == 2)
		for (pv = cur_vars; ; ++pv)
			if (!pv->name)
				{
				fprintf(stderr,
					"read_vars: ignoring unknown parameter: \"%s\"\n", name);
				break;
				}
			else if (!strcmp(pv->name, name))
				{
				pv->data = val;
				break;
				}
	fclose(file);
	return (NULL);
	}

/* publish our pid, delete file at termination
 */
char *adm_pid_file(char *fn)
	{
	static char *cur_fn = NULL;
	FILE *file;
	unsigned int pid;

	if (fn)
		{
		if (!(file = fopen(fn, "r")))
			;	/* no file */
		else if (fscanf(file, "%d", &pid) != 1)
			;	/* bad format */
		else
			{
			kill(pid, SIGQUIT);	/* remove old daemon */
			waitpid(pid, NULL, 0); /* wait for old daemon to exit */
			}
		if (file)
			fclose(file);
		if (!(file = fopen(fn, "w")))
			return ("can't open daemon pid file");
		else if (!fprintf(file, "%d\n", getpid()))
			return ("can't write pid file");
		else
			cur_fn = fn;
		if (file)
			fclose(file);
		}
	else if (!cur_fn)
		;
	else if (!unlink(cur_fn))
		return ("can't unlink pid file");
	else
		cur_fn = NULL;
	return (NULL);
	}

/* pack labelled bottons into vertical box
 */
GtkWidget *bt_entry(GtkWidget *vb, char *name)
	{
	GtkWidget *hb, *checkbutton;

	hb = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vb), hb, FALSE, TRUE, 10);
	checkbutton = gtk_check_button_new_with_label(name);
	gtk_widget_set_usize(checkbutton, 200, 20);
	gtk_box_pack_start(GTK_BOX(hb), checkbutton, FALSE, TRUE, 5);
	return (hb);
	}

/* generic delete function for a pop-up window
 */
gint popup_delete(GtkWidget *widget, GtkWidget *event, gpointer data)
	{
	gtk_widget_destroy((GtkWidget *)data);
	return (TRUE);
	}

/* generic dismiss function for a pop-up window
 */
gint popup_close(GtkButton *button, gpointer data)
	{
	gtk_widget_destroy((GtkWidget *)data);
	return (TRUE);
	}

void info_popup(char *info)
	{
	GtkWidget *window, *label, *vb, *checkbutton;

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), LC_TITLE);					// hitachi add
	gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);	// hitachi add
	gtk_signal_connect(GTK_OBJECT (window), "delete_event",
			GTK_SIGNAL_FUNC(popup_delete), window);
	gtk_container_set_border_width(GTK_CONTAINER (window), 10);
	vb = gtk_vbox_new(FALSE, 10);
	gtk_container_set_border_width(GTK_CONTAINER(vb), 10);
	gtk_container_add(GTK_CONTAINER(window), vb);
	label = gtk_label_new(info);
	gtk_box_pack_start(GTK_BOX(vb), label, FALSE, FALSE, 5);
	checkbutton = gtk_button_new_with_label(LC_OK);						// hitachi mod
	gtk_signal_connect(GTK_OBJECT(checkbutton), "clicked",
		(GtkSignalFunc)popup_close, GTK_OBJECT(window));
	gtk_box_pack_start(GTK_BOX(vb), checkbutton, FALSE, FALSE, 5);
	gtk_widget_show_all(window);
	}

gint pb_popup_delete(GtkWidget *widget, GtkWidget *event, gpointer data)
	{
	gtk_widget_destroy((GtkWidget *)data);
	return (TRUE);
	}

/* powerdown selected from power button pop-up window
 */
gint pb_powerdown(GtkButton *button, gpointer data)
	{
	gtk_widget_destroy((GtkWidget *)data);
	powerdown();
	pb_popup_nest = 0;
	return (TRUE);
	}

/* power saving mode selected from power button pop-up window
 */
gint pb_powersave(GtkButton *button, gpointer data)
	{
	gtk_widget_destroy((GtkWidget *)data);
	powersave();
	pb_popup_nest = 0;
	return (TRUE);
	}

/* dismiss power button pop-up window
 */
gint pb_popup_close(GtkButton *button, gpointer data)
	{
	gtk_widget_destroy((GtkWidget *)data);
	pb_popup_nest = 0;
	return (TRUE);
	}

/* power button pressed pop-up window
 */
void pb_popup(char *info)
	{
	GtkWidget *window, *label, *vb, *checkbutton, *frame;

	if (pb_popup_nest)
		return;
	pb_popup_nest = 1;
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), LC_TITLE);					// hitachi add
	gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);	// hitachi add
	gtk_signal_connect(GTK_OBJECT (window), "delete_event",
		GTK_SIGNAL_FUNC(popup_delete), window);
	gtk_container_set_border_width(GTK_CONTAINER (window), 10);
	frame = gtk_frame_new(info);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 10);
	gtk_container_add(GTK_CONTAINER(window), frame);
	vb = gtk_vbox_new(FALSE, 10);
	gtk_container_set_border_width(GTK_CONTAINER(vb), 10);
	gtk_container_add(GTK_CONTAINER(frame), vb);
	checkbutton = gtk_button_new_with_label(LC_ENTER_LOW_POWER_MODE);
	gtk_signal_connect(GTK_OBJECT(checkbutton), "clicked",
		(GtkSignalFunc)pb_powersave, GTK_OBJECT(window));
	gtk_box_pack_start(GTK_BOX(vb), checkbutton, FALSE, FALSE, 5);
	checkbutton = gtk_button_new_with_label(LC_SHUTDOWN_SYSTEM);
	gtk_signal_connect(GTK_OBJECT(checkbutton), "clicked",
		(GtkSignalFunc)pb_powerdown, GTK_OBJECT(window));
	gtk_box_pack_start(GTK_BOX(vb), checkbutton, FALSE, FALSE, 5);
	checkbutton = gtk_button_new_with_label(LC_DISMISS);
	gtk_signal_connect(GTK_OBJECT(checkbutton), "clicked",
		(GtkSignalFunc)pb_popup_close, GTK_OBJECT(window));
	gtk_box_pack_start(GTK_BOX(vb), checkbutton, FALSE, FALSE, 5);
	gtk_widget_show_all(window);
	}

/* power button pressed event
 */
void power_button()
	{
	extern int bl_state;

	g_print("");
	switch (cur_vars[PMVAR_MD].data)
		{
	case PBM_PU:
		set_bl_state(BL_USREVENT, 0);
		pb_popup(LC_POWER_BUTTON_PRESSED);
		break;
	case PBM_PS:
		if (bl_state == BL_OFFEWAIT)	/* effectively toggles backlight */
			set_bl_state(BL_USREVENT, 0);
		else
			powersave();
		break;
	case PBM_PD:
		powerdown();
		break;
		}
	}

/* Enter power saving mode as selected by the user.  This is a placeholder
 * for future PCI/USB/PCMCIA power management.  Note we may enter here
 * from a time out or from the power button handler.  There is no
 * differentiation currently.
 */
int powersave()
	{
	set_bl_state(BL_SETTLE, 0);
	}

/* Power down.  Note we may enter here from a time out or from the power
 * button handler.  There is no differentiation currently.
 */
int powerdown()
	{
	set_bl_state(BL_SHUTDOWN, 0);
#ifdef PROTOTYPE
	printf("shutdown\n");
#else
	system(PMHALT_PATH);
#endif
	}

gint shutdown_delete(GtkWidget *widget, GtkWidget *event, gpointer data)
	{
	gtk_widget_destroy((GtkWidget *)data);
	lowbatt_warning = 0;
	lobatt_secs = 0;
	return (TRUE);
	}

gint shutdown_close(GtkButton *button, gpointer data)
	{
	gtk_widget_destroy((GtkWidget *)data);
	lowbatt_warning = 0;
	lobatt_secs = 0;
	return (TRUE);
	}

gint shutdown_now(GtkButton *button, gpointer data)
	{
	gtk_widget_destroy((GtkWidget *)data);
	lowbatt_warning = 0;
	lobatt_secs = 0;
	powerdown();
	return (TRUE);
	}

void shutdown_popup(char *info)
	{
	GtkWidget *window, *label, *vb, *checkbutton;

	if (lowbatt_warning)
		return;
	lowbatt_warning = 1;
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), LC_TITLE);					// hitachi add
	gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);	// hitachi add
	gtk_signal_connect(GTK_OBJECT(window), "delete_event",
			GTK_SIGNAL_FUNC(shutdown_delete), window);
	gtk_container_set_border_width(GTK_CONTAINER(window), 10);
	vb = gtk_vbox_new(FALSE, 10);
	gtk_container_set_border_width(GTK_CONTAINER(vb), 10);
	gtk_container_add(GTK_CONTAINER(window), vb);
	label = gtk_label_new(info);
	gtk_box_pack_start(GTK_BOX(vb), label, FALSE, FALSE, 5);
	if (!cur_vars[PMVAR_AC].data)
		{
		checkbutton = gtk_button_new_with_label(LC_SHUTDOWN_NOW);
		gtk_signal_connect(GTK_OBJECT(checkbutton), "clicked",
			(GtkSignalFunc)shutdown_now, GTK_OBJECT(window));
		gtk_box_pack_start(GTK_BOX(vb), checkbutton, FALSE, FALSE, 5);
		}
	else
		{
		label = gtk_label_new(LC_SHUTTING_DOWN_SYSTEM);
		gtk_box_pack_start(GTK_BOX(vb), label, FALSE, FALSE, 5);
		}
	checkbutton = gtk_button_new_with_label(LC_DISMISS);
	gtk_signal_connect(GTK_OBJECT(checkbutton), "clicked",
		(GtkSignalFunc)shutdown_close, GTK_OBJECT(window));
	gtk_box_pack_start(GTK_BOX(vb), checkbutton, FALSE, FALSE, 5);
	gtk_widget_show_all(window);
	}

gint power_monitor_poll(gpointer p)
	{
	power_monitor(0);
	return (TRUE);
	}

/* monitor power state
 */
void power_monitor(int sig)
	{
	extern int bl_state;
	static int err_cnt = {0};
	static int pbp_old = {0};
	GtkAdjustment *adj;
	int value, battery_level, ac_input, pbp, battery_present;
	char name[100];	/* see fscanf %s arg below */
	FILE *file;
	char *err;

	if (bl_state == BL_SHUTDOWN)
		return;	/* don't rattle around if shutting down */
	if (!sig)
		++lobatt_secs;
	pbp = 0;
	err = NULL;
	if (!(file = fopen(POWER_STATUS_PATH, "r")))
		err = "open";
	else if (flock(file->_fileno, LOCK_SH|LOCK_NB))
		err = "lock";
	else
		while (fscanf(file, "%99s %d", name, &value) == 2)
			if (!strcmp("battery_level", name))
				battery_level = value;
			else if (!strcmp("ac_input", name))
				ac_input = value;
			else if (!strcmp("powerbutton_pressed", name))
				pbp = value;
			else if (!strcmp("battery_present", name))
				battery_present = value;
	if (file)
		{
		flock(file->_fileno, LOCK_UN);
		fclose(file);
		}
	if (!err)
		{
		err_cnt = 0;
		if (ac_input || LO_BATT_TRIP < battery_level)
			;
		else if (lobatt_secs < battery_level)
			;
		else if (battery_present)
			{
			if (cur_vars[PMVAR_AU].data)
				g_print("");
			if (cur_vars[PMVAR_PU].data)
				shutdown_popup(LC_LOW_BATTERYX);
			if (cur_vars[PMVAR_AC].data)
				powerdown();
			}
		if (pbp && pbp_old != pbp)
			power_button(), pbp_old = pbp;
		}
	else if (++err_cnt < 3)
		;
	else
		fprintf(stderr, "Daemon status file error: %s\n", err);
	}

void reload_conf()
	{
	read_vars(PMCONFIG_PATH);
	if (!cur_vars[PMVAR_BLE].data || !cur_vars[PMVAR_BL].data)
		set_bl_state(BL_UNTIMED_ON, 0);
	else
		set_bl_state(BL_TIMED_ON, cur_vars[PMVAR_BL].data);
	}

void reload_conf_notify()
	{
	char buf[100];

	reload_conf();
	sprintf(buf, "%s\n" LC_RELOADED_CONFIGURATION "\n", procname);
	info_popup(buf);
	}

void sig_exit(int sig)
	{
	cleanup(sig);
	}

int daemon_ctl(int cmd)
	{
	static int pmacpid_pid = {0};

	switch (cmd)
		{
	case DAEMON_DELETE:
		if (kill(pmacpid_pid, SIGINT) < 0)
			perror("couldn't kill "PMACPID_PATH);
		break;
	case DAEMON_START:
		if (pmacpid_pid = fork())
			;
		else if (pmacpid_pid < 0)
			{
			perror("couldn't fork");
			exit(12);
			}
		else if (execl(PMACPID_PATH, PMACPID_PATH, NULL) < 0)
			{
			perror("couldn't exec: "PMACPID_PATH);
			exit(13);
			}
		break;
	default:
		fprintf(stderr, "daemon_ctl() unknown cmd %d\n", cmd);
		}
	return (0);
	}

/* tidy-up before exiting
 * NB: if terminating due to a SIGQUIT, do not remove our PMD_PID_PATH
 * to avoid racing with the new daemon which is creating this file
 */
void cleanup(int sig)
	{
	if (sig != SIGQUIT)
		adm_pid_file(NULL);
	daemon_ctl(DAEMON_DELETE);
	_exit(0);
	}
	
/* start pmd and support daemons, read pm variables
 */
int main(int argc, char *argv[])
	{
	char *msg;
	
	procname = argv[0];
	on_exit(cleanup, NULL);
	if (msg = adm_pid_file(PMD_PID_PATH))
		g_print("adm_pid_file: %s\n", msg);
	signal(SIGHUP, reload_conf_notify);
	signal(SIGUSR1, power_monitor);
	signal(SIGUSR2, dump_vars);
	signal(SIGQUIT, sig_exit);
	signal(SIGINT, sig_exit);
	signal(SIGTERM, sig_exit);
	gtk_set_locale();
	gtk_init(&argc, &argv);
	daemon_ctl(DAEMON_START);
	set_bl_state(BL_INIT, 0);
	init_vars();
	read_vars(PMCONFIG_PATH);
	reload_conf();
	gtk_timeout_add(BATT_POLL_INTV, power_monitor_poll, NULL); 
	gtk_main();
	return (0);
	}
